
class Dices(nb_values:Int, val nb_faces:Int) {

  // Face values, not very abstractly...
  object Face {
    type Face = Int
    val min = 1
    val max = nb_values
    val values = min to max
  }

  // A dice, described as a vector of face values
  class Dice(protected val faces:Vector[Int]) extends Ordered[Dice] {

    assert (faces.length == nb_faces)
    assert (faces.forall { x => Face.min <= x && x <= Face.max })

    def apply(i:Int) = faces(i)
    override def toString() = faces.mkString

    // Lexicographic ordering
    def compare(that:Dice) : Int = {
      for (i <- 1 to nb_faces) {
        val c = this.faces(i).compare(that.faces(i))
        if (c!=0) { return c }
      }
      return 0
    }

    // Count how many times this is better than that,
    // among all possible outcomes of rolling those dices.
    // XXX Relies on an unspecified assumption that faces are
    // sorted in increasing order.
    def scoreAgainst(that:Dice) : Int = {
      var i,j = nb_faces-1
      var score = 0
      while (i>=0 && j>=0) {
        if (this(i)>that(j)) {
          score += i+1
          i -= 1
        } else {
          j -= 1
        }
      }
      score
    }

    // Score ordering
    def >>(that:Dice) : Boolean = {
      scoreAgainst(that)*2>(nb_faces*nb_faces)
    }

  }

  // Iterator of all dices in lexicographic order
  // e.g. 111, 112, 113, 122, 123, 133, 222...
  class Dices(init:Option[Dice]) extends Iterator[Dice] {

    def this() = this(None);
    def this(d:Dice) = this(Some(d));
    
    val cur = new Array[Face.Face](nb_faces)
    var i = 0
    init match {
      case None =>
        // Start with cur full of zeroes:
        // not a proper dice but its successor will be valid
      case Some(d) =>
        // Start with cur = d, after which we will
        // enumerate the strictly greater dices.
        for (i <- 0 until nb_faces)
          cur(i) = d(i)
        i = nb_faces-1
        while (i>=0 && cur(i)==Face.max) i=i-1
    }

    def hasNext = i>=0 // cur exists (_!=Face.max)
    def next = {
      if (cur(0)==0) {
        for (i <- 0 until nb_faces) cur(i) = Face.min
        i = nb_faces-1
      } else {
        // Increase last non-maximal face
        assert (cur(i)!=Face.max)
        cur(i) = cur(i)+1
        // Propagate
        for (j <- i+1 until nb_faces) cur(j)=cur(i)
        // Update index of last non-maximal face
        if (cur(i)<Face.max)
          i=nb_faces-1
        else
          while (i>=0 && cur(i)==Face.max) i=i-1
      }
      val v = Vector(cur: _*)
      new Dice(v)
    }

  }

  // Same thing in a more functional style
  def iter(cur:Int,faces:Int) : Iterator[List[Int]] = {
    assert (Face.min <= cur && cur <= Face.max)
    if (faces==0)
      Iterator(List()) // singleton iterator
    else
      if (cur==Face.max) {
        (for { d <- iter(cur,faces-1) } yield (cur::d))
      } else {
        (for { d <- iter(cur,faces-1) } yield (cur::d))
        (for { d <- iter(cur+1,faces) } yield d)
      }
  }

  // For testing purposes: count number of dices,
  // without enumerating
  def count(cur:Int,faces:Int) : Int = {
    if (faces==0) 1 else
      if (cur==Face.max) count(cur,faces-1) else
        count(cur,faces-1)+count(cur+1,faces)
  }
  def count() : Int = count(Face.min,nb_faces)

}

object Main extends App {

  // Expected numbers of results
  // 4,6 : 0
  // 5,6 : 15
  // 6,6 : 581
  val n = 5
  println("Finding weird triples with "+n+" values on 6 faces...")
  val Dices = new Dices(n,6)

  // Trouver 3 dés a, b et c tels que
  // a >> b >> c >> a
  // a < b, a < c
  var count = 0
  for {
    a <- new Dices.Dices
    b <- new Dices.Dices(a)
    if a >> b
      c <- new Dices.Dices(a)
    if (b >> c && c >> a)
  } {
    assert (a >> b && b >> c && c >> a)
    assert (!(b >> a || c >> b || a >> c))
    count += 1
    println(count + ": " + a + " " + b + " " + c)
  }

}
